#include "mem.h"
#include "list.h"
#include "vga.h"

static struct mem_desc *page_desc = (struct mem_desc *)&__ld_kernel_end;
static struct mem_desc *free_page;
static struct mem_desc *used_page;
static unsigned long bottom_mem;
static unsigned long top_mem;


int memSetup(unsigned long upper_mem)
{
	printf("Free Mem going from %d to %d\n", &__ld_kernel_end, upper_mem * 1024);
	unsigned long memdesc_end =
	    (unsigned long)page_desc +
	    ((upper_mem) / (PAGE_SIZE / 1024)) * sizeof(struct mem_desc);
	uint lastUsed = (memdesc_end >> PAGE_SHIFT) + 1;
	list_init(free_page);
	list_init(used_page);
	bottom_mem = lastUsed;
	top_mem = upper_mem;
	for (uint i = 0; i < (upper_mem / (PAGE_SIZE / 1024)); i++) {
		struct mem_desc *mem = &page_desc[i];
		if (i < lastUsed) {
			mem->ref = 1;
			list_add_tail(used_page, mem);
		} else {
			mem->ref = 0;
			list_add_tail(free_page, mem);
		}
		mem->phy_addr = i * PAGE_SIZE;
	}
	return 0;
}

struct mem_desc *addr2memDesc(unsigned long addr)
{
	if (addr > top_mem || addr < bottom_mem)
		return NULL;
	int idx = addr >> PAGE_SHIFT;
	return page_desc + idx;
}


unsigned long allocPage(void)
{
	if (list_is_empty(free_page))
		return (unsigned long)NULL;
	struct mem_desc *mem = list_pop_head(free_page);
	mem->ref             = 1;
	list_add_tail(used_page, mem);
	return mem->phy_addr;
}

int unrefPage(unsigned long addr)
{
	struct mem_desc *mem = addr2memDesc(addr);
	if (!mem) {
		return -1;
	}
	mem->ref--;
	if (mem->ref == 0) {
		list_add_tail(free_page, mem);
	}

	return 0;
}
